{=  Last modified                                                       =}
{=  $Date: 09.05.98 - 13:23:05 $                                      =}
unit Socks;


interface
uses Winsock, Classes, SysUtils, messg;

type

  ESocketErrorException = class(Exception);
  ESocketTimeoutException = class(Exception);
  PLongint = ^Longint;

  TUDPSocket= class(TObject)
  private
    sock:TSocket;
  public
    Constructor Create(port:string);
    function SendBroadcast(var buf;len:integer;port:string):integer;
    function SendToAddr(var buf;len:integer;adr,port:string):integer;
    function ReceiveFrom(var buf;len:integer;var rip,rport,rname:string):integer;
  end;


  TClientSocket = class(TObject)
  private
    portn: Word;
    peerip: string;
    peername: string;
  protected
    sock: tSocket;
  public
    constructor Create;
    destructor Destroy;override;
    procedure Connect(adr: string; PORT: string);
    procedure WaitForData;
    procedure WaitForDataTime(A: Word);
    procedure Disconnect;
    function read (var Buf; len: Word): Word;
    function write (var Buf; len: Word): Word;
  end;

  TServerSocket = class(TObject)
  protected
    sock: tSocket;
  private
    portn: Word;
    peerip: string;
    peername: string;
  public
    property FPeerIP: string read peerip;
    property FPeerName: string read peername;
    constructor Create(PORT: string);
    destructor Destroy;override;
    procedure Listen;
    procedure WaitForConnection;
    function AcceptConnection: TClientSocket;
    procedure Disconnect;
  end;

var
  WSAReady: Boolean = False;


implementation

function GetPortByname(port,proto:string):integer;
var
 PSE:PServEnt;
begin
 PSE := getservbyname(PChar(PORT), PChar(proto));
 if PSE = nil then begin
   Warn(Format('cannot resolve service name %s; usign it as port number',[PORT]));
   Result:=StrToInt(PORT);
   Exit;
 end
 else begin
   Result := ntohs(PSE^.s_port);
   Deb(Format('resolved; port number %d',[ntohs(PSE^.s_port)]), 2);
 end;
end;

procedure SocketError(const Msg: string);
var S: string;
begin
  S := Format('%s; error code %d',[Msg, WSAGetLastError]);
  ERR(S);
  raise ESocketErrorException.Create(S);
end;

var
  WSAData: TWSAData;

procedure StartWSA;
begin
  Deb('Starting WSA... ', 1);
  if WSAStartUp(257, WSAData) <> 0 then SocketError('in socks.StartWSA: WSAStartUp() failed');
  WSAReady := True;
  Deb('WSA: ' + WSAData.szDescription + ' ready', 1);
end;

{----------------------------- Server Socket implementation ---------------}

constructor TServerSocket.Create(PORT: string);
var
  sn : TSockAddrIn;
  PSE: PServEnt;   
begin
  Assert(WSAReady, 'WSA not initialized');
  inherited Create;
  sock := INVALID_SOCKET; peername := ''; peerip := '';
  Deb(Format('Creating master socket on port %s',[PORT]), 1);
  sock := socket(PF_INET, SOCK_STREAM, getprotobyname('tcp').p_proto);
  if sock = INVALID_SOCKET then SocketError('in TServerSocket.Create: socket() failed');
  Deb(Format('master socket: %d',[sock]), 2);
  sn.sin_family := AF_INET;
  sn.sin_addr.s_addr := INADDR_ANY;
  Deb(Format('resolving port name %s',[PORT]), 2);
  PSE := getservbyname(PChar(PORT), PChar('tcp'));
  if PSE = nil then begin
    Warn(Format('cannot resolv; usign port number %s',[PORT]));
    sn.sin_port := htons(StrToInt(PORT));
    portn := StrToInt(PORT);
  end
  else begin
    sn.sin_port := PSE^.s_port;
    portn := ntohs(PSE^.s_port);
    Deb(Format('resolved; port number %d',[portn]), 2);
  end;
  Deb('Binding socket', 2);
  if bind(sock, sn, sizeof(sn)) <> 0 then SocketError('in TServerSocket.Create: bind() failed');
end;

destructor TServerSocket.Destroy;
begin
  if sock <> INVALID_SOCKET then Disconnect;
  inherited Destroy;
end;


procedure TServerSocket.Listen;
begin
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb(Format('Putting socket on port %d in listen state',[portn]), 1);
  if Winsock.Listen(sock, 5) <> 0 then SocketError('in TServerSocket.Listen: listen() failed');
end;

procedure TServerSocket.WaitForConnection;
var
  fdset: TFDset;
begin
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb('Waiting for connection...', 1);
  fdset.fd_count := 1;
  fdset.fd_array[0] := sock;
  if Select(0, @fdset, nil, nil, nil) = SOCKET_ERROR then SocketError('in TServerSocket.WaitForConnect: select() failed');
end;

function TServerSocket.AcceptConnection;
var
  TmpSock: tSocket;      
  sad    : TSockAddr;
  I      : Integer;      
  phe    : PHostEnt;     
  TmpCs  : TClientSocket;
begin
  Result := nil;
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb('Accepting connection...', 1);
  I := sizeof(TSockAddr);
  TmpSock := Accept(sock, @sad, @I);
  if TmpSock = INVALID_SOCKET then SocketError('in TServerSocket.AcceptConnection: accept() failed');
  peerip := IntToStr(Ord(sad.sa_data[2])) + '.' + IntToStr(Ord(sad.sa_data[3])) + '.' +
    IntToStr(Ord(sad.sa_data[4])) + '.' + IntToStr(Ord(sad.sa_data[5]));
  phe := gethostbyaddr(@(sad.sa_data[2]), 4, PF_INET);
  if phe = nil then begin
    Warn(Format('in TServerSocket.AcceptConnection: gethostbyaddr() failed; error code %d',[WSAGetLastError]));
    peername := '';
  end
  else peername := phe^.h_name;
  TmpCs := TClientSocket.Create;
  TmpCs.sock := TmpSock;
  Result := TmpCs;
end;

procedure TServerSocket.Disconnect;
begin
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb(Format('Closing connection on socket %d...',[sock]), 1);
  if closesocket(sock) = SOCKET_ERROR then SocketError('in TClientSocket.disconnect: closesocket() failed');
  sock := INVALID_SOCKET;
  peername := '';
end;



{--------------------- End of Server Socket inplementation ---------------}



{--------------------- Client Socket implementation ----------------------}

constructor TClientSocket.Create;
begin
  Assert(WSAReady, 'WSA not initialized');
  inherited Create;
  sock := INVALID_SOCKET; peername := ''; peerip := '';
end;


procedure TClientSocket.Connect(adr: string; PORT: string);
var phe: PHostEnt;
  sin   : TSockAddrIn;
  PSE   : PServEnt;
  szData: PChar;      
begin
  Assert(WSAReady, 'WSA not initialized');
  Deb('Creating client socket...', 1);
  sock := socket(PF_INET, SOCK_STREAM, getprotobyname('tcp').p_proto);
  if sock = INVALID_SOCKET then SocketError('in TClientSocket socket() failed');
  sin.sin_family := AF_INET;
  PSE := getservbyname(PChar(PORT), PChar('tcp'));
  if PSE = nil then begin
    Warn(Format('cannot resolve service name %s; usign is as port number',[PORT]));
    sin.sin_port := htons(StrToInt(PORT));
    portn := StrToInt(PORT);
  end
  else begin
    sin.sin_port := PSE^.s_port;
    portn := ntohs(PSE^.s_port);
    Deb(Format('resolved; port number %d',[portn]), 2);
  end;
  szData := PChar(adr);
  sin.sin_addr.s_addr := inet_addr(szData);
  if sin.sin_addr.s_addr = INADDR_NONE then begin
    Deb(Format('%s does not look like ip. Trying to use it as host name ',[adr]), 2);
    phe := gethostbyname(szData);
    if phe = nil then SocketError(Format('Host name %s unknown',[adr]));
    Deb('resolved host name ', 2);
    sin.sin_addr.s_addr := Longint(PLongint(phe^.h_addr_list^)^);
  end;
  Deb(Format('trying to connect to %s:%s',[adr, PORT]), 1);
  if Winsock.Connect(sock, sin, sizeof(sin)) <> 0 then SocketError(Format('Cannot connect to %s:%s',[adr, PORT]));
  peername := adr;
end;

procedure TClientSocket.WaitForData;
var
  fdset: TFDset;
begin
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb(Format('Waiting for data on socket %d',[sock]), 1);
  fdset.fd_count := 1;
  fdset.fd_array[0] := sock;
  if Select(0, @fdset, nil, nil, nil) = SOCKET_ERROR then SocketError('in TClientSocket.WaitForData: select() failed');
end;


procedure TClientSocket.WaitForDataTime;
var
  fdset: TFDset;  
  TM   : TTimeVal;
  Res  : Integer; 
begin
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb(Format('Waiting for data on socket %d %d microseconds',[sock, A]), 1);
  TM.tv_sec := 0;
  TM.tv_usec := A;
  fdset.fd_count := 1;
  fdset.fd_array[0] := sock;
  Res := Select(0, @fdset, nil, nil, @TM);
  if Res = 0 then raise ESocketTimeoutException.Create('');
  if Res = SOCKET_ERROR then SocketError('in TClientSocket.WaitForDataTime: select() failed');
end;


function TClientSocket.read;
begin
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb(Format('Reading data on socket %d...',[sock]), 1);
  Result := recv(sock, Buf, len, 0);
  if Result = SOCKET_ERROR then SocketError('in TClientSocket.read: recv() failed');
end;

function TClientSocket.write;
begin
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb(Format('Writing data to socket %d...',[sock]), 1);
  Result := send(sock, Buf, len, 0);
  if Result = SOCKET_ERROR then SocketError('in TClientSocket.write: send() failed');
end;


procedure TClientSocket.Disconnect;
begin
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb(Format('Closing connection on socket %d...',[sock]), 1);
  if closesocket(sock) = SOCKET_ERROR then SocketError('in TClientSocket.disconnect: closesocket() failed');
  sock := INVALID_SOCKET;
  peername := '';
end;


destructor TClientSocket.Destroy;
begin
  if sock <> INVALID_SOCKET then Disconnect;
  inherited Destroy;
end;
{------------------- End of TClientSocket implementation--------------}



{------------------- Begin TUDPSocket implementation ----------------}
constructor TUDPSocket.Create;
var sin:TSockAddrIN;
begin
 Assert(WSAReady, 'WSA not initialized');
 inherited Create;
 Deb('Creating UDP socket',1);
 sock:=INVALID_SOCKET;
 sock:=socket(PF_INET,SOCK_DGRAM,getprotobyname('udp').p_proto);
 if sock=INVALID_SOCKET then SocketError('in TUDPSocket.Create socket() failed');
 sin.sin_family := AF_INET;
 sin.sin_addr.s_addr := INADDR_ANY;
 sin.sin_port := htons(GetPortByName(PORT,'udp'));
 if bind(sock, sin, sizeof(sin))= SOCKET_ERROR then SocketError('in TUDPSocket.Create: bind() failed');
end;

function TUDPSocket.SendBroadcast(var buf;len:integer;port:string):integer;
var i,si:integer;
    sin:TSockAddrIn;
begin
  si:=SizeOf(integer);
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb(Format('Sending broadcast on UDP socket %d...',[sock]), 1);
  if getsockopt(sock,SOL_SOCKET,SO_BROADCAST,@i,si)=SOCKET_ERROR then SocketError('in TUDPSocket.SendBroadcast getsockopt() failed');
  if i=0 then begin
    i:=1;
    if setsockopt(sock,SOL_SOCKET,SO_BROADCAST,@i,si)=SOCKET_ERROR then SocketError('in TUDPSocket.SendBroadcast setsockopt() failed');
  end;
 sin.sin_family:=AF_INET;
 sin.sin_addr.s_addr:=INADDR_BROADCAST;
 sin.sin_port:=htons(GetPortByName(port,'udp'));
 Result:=sendto(sock,buf,len,0,sin,SizeOf(sin));
 if Result=SOCKET_ERROR then SocketError('in TUDPSocket.SendBroadCast sendto() failed');
end;


function TUDPSocket.SendToAddr(var buf;len:integer;adr,port:string):integer;
var i,si:integer;
    sin:TSockAddrIn;
    szData:Pchar;
    phe:PHostEnt;
begin
  si:=SizeOf(integer);
  Assert(WSAReady, 'WSA not initialized');
  Assert(sock <> INVALID_SOCKET, 'Socket operation on non socket');
  szData := PChar(adr);
  sin.sin_addr.s_addr := inet_addr(szData);
  if sin.sin_addr.s_addr = INADDR_NONE then begin
    Deb(Format('%s does not look like ip. Trying to use it as host name ',[adr]), 2);
    phe := gethostbyname(szData);
    if phe = nil then SocketError(Format('Host name %s unknown',[adr]));
    Deb('resolved host name ', 2);
    sin.sin_addr.s_addr := Longint(PLongint(phe^.h_addr_list^)^);
  end;
  Deb(Format('Sending broadcast on UDP socket %s:%s on %d...',[adr,port,sock]), 1);
  if getsockopt(sock,SOL_SOCKET,SO_BROADCAST,@i,si)=SOCKET_ERROR then SocketError('in TUDPSocket.SendBroadcast getsockopt() failed');
  if i=0 then begin
    i:=1;
    if setsockopt(sock,SOL_SOCKET,SO_BROADCAST,@i,si)=SOCKET_ERROR then SocketError('in TUDPSocket.SendBroadcast setsockopt() failed');
  end;
 sin.sin_family:=AF_INET;
 sin.sin_addr.s_addr:=INADDR_BROADCAST;
 sin.sin_port:=htons(GetPortByName(port,'udp'));
 Result:=sendto(sock,buf,len,0,sin,SizeOf(sin));
 if Result=SOCKET_ERROR then SocketError('in TUDPSocket.SendBroadCast sendto() failed');
end;

function TUDPSocket.ReceiveFrom(var buf;len:integer;var rip,rport,rname:string):integer;
var sin:TSockAddrIN;
    ssin:integer;
    phe:PHostEnt;
begin
  ssin:=SizeOf(TSockAddrIn);
  Assert(WSAReady, 'WSA not initialized');
  Assert(Self.sock <> INVALID_SOCKET, 'Socket operation on non socket');
  Deb(Format('Receiving data on socket %d...',[sock]),1);
  Result:=winsock.recvfrom(sock,buf,len,0,sin,ssin);
  if Result=SOCKET_ERROR then SocketError('in TUDPSocket.ReceiveFrom recvfrom() failed');
  rip := IntToStr(Ord(sin.sa_data[2])) + '.' + IntToStr(Ord(sin.sa_data[3])) + '.' +
    IntToStr(Ord(sin.sa_data[4])) + '.' + IntToStr(Ord(sin.sa_data[5]));
  phe:=gethostbyaddr(@sin.sa_data[2],4,PF_INET);
  if phe=nil then Warn('in TUDPSocket.ReceiveFrom gethostbyaddr() failed');
  if phe<>nil then rname:=phe.h_name;
  rport:='';
  Deb(Format('After proc sock is %d',[sock]),1);
end;

initialization
  StartWSA;

finalization
  if WSACleanup=SOCKET_ERROR then SocketError('in finalization WSACleanup() failed');

end.

